"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class RenderingControl3ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:RenderingControl:3' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'RenderingControl3'

    SERVICE_DEFAULT_VARIABLES = {
        "AllowedTransformSettings": { "data_type": "string", "default": None, "allowed_list": None},
        "BlueVideoBlackLevel": { "data_type": "ui2", "default": None, "allowed_list": None},
        "BlueVideoGain": { "data_type": "ui2", "default": None, "allowed_list": None},
        "Brightness": { "data_type": "ui2", "default": None, "allowed_list": None},
        "ColorTemperature": { "data_type": "ui2", "default": None, "allowed_list": None},
        "Contrast": { "data_type": "ui2", "default": None, "allowed_list": None},
        "GreenVideoBlackLevel": { "data_type": "ui2", "default": None, "allowed_list": None},
        "GreenVideoGain": { "data_type": "ui2", "default": None, "allowed_list": None},
        "HorizontalKeystone": { "data_type": "i2", "default": None, "allowed_list": None},
        "Loudness": { "data_type": "boolean", "default": None, "allowed_list": None},
        "Mute": { "data_type": "boolean", "default": None, "allowed_list": None},
        "PresetNameList": { "data_type": "string", "default": None, "allowed_list": None},
        "RedVideoBlackLevel": { "data_type": "ui2", "default": None, "allowed_list": None},
        "RedVideoGain": { "data_type": "ui2", "default": None, "allowed_list": None},
        "Sharpness": { "data_type": "ui2", "default": None, "allowed_list": None},
        "TransformSettings": { "data_type": "string", "default": None, "allowed_list": None},
        "VerticalKeystone": { "data_type": "i2", "default": None, "allowed_list": None},
        "Volume": { "data_type": "ui2", "default": None, "allowed_list": None},
        "VolumeDB": { "data_type": "i2", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "AllowedDefaultTransformSettings": { "data_type": "string", "default": None, "allowed_list": None},
        "DefaultTransformSettings": { "data_type": "string", "default": None, "allowed_list": None},
        "LastChange": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetAllAvailableTransforms(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAllAvailableTransforms action.

            :returns: "AllAllowedTransformSettings"
        """
        arguments = { }

        out_params = self.call_action("GetAllAvailableTransforms", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AllAllowedTransformSettings",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAllowedDefaultTransforms(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAllowedDefaultTransforms action.

            :returns: "AllowedDefaultTransformSettings"
        """
        arguments = { }

        out_params = self.call_action("GetAllowedDefaultTransforms", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AllowedDefaultTransformSettings",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAllowedTransforms(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAllowedTransforms action.

            :returns: "CurrentAllowedTransformSettings"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetAllowedTransforms", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentAllowedTransformSettings",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBlueVideoBlackLevel(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBlueVideoBlackLevel action.

            :returns: "CurrentBlueVideoBlackLevel"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetBlueVideoBlackLevel", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentBlueVideoBlackLevel",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBlueVideoGain(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBlueVideoGain action.

            :returns: "CurrentBlueVideoGain"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetBlueVideoGain", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentBlueVideoGain",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBrightness(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBrightness action.

            :returns: "CurrentBrightness"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetBrightness", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentBrightness",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetColorTemperature(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetColorTemperature action.

            :returns: "CurrentColorTemperature"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetColorTemperature", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentColorTemperature",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetContrast(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetContrast action.

            :returns: "CurrentContrast"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetContrast", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentContrast",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDefaultTransforms(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDefaultTransforms action.

            :returns: "CurrentDefaultTransformSettings"
        """
        arguments = { }

        out_params = self.call_action("GetDefaultTransforms", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentDefaultTransformSettings",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetGreenVideoBlackLevel(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetGreenVideoBlackLevel action.

            :returns: "CurrentGreenVideoBlackLevel"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetGreenVideoBlackLevel", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentGreenVideoBlackLevel",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetGreenVideoGain(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetGreenVideoGain action.

            :returns: "CurrentGreenVideoGain"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetGreenVideoGain", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentGreenVideoGain",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetHorizontalKeystone(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetHorizontalKeystone action.

            :returns: "CurrentHorizontalKeystone"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetHorizontalKeystone", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentHorizontalKeystone",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetLoudness(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLoudness action.

            :returns: "CurrentLoudness"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetLoudness", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentLoudness",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMute(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMute action.

            :returns: "CurrentMute"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetMute", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentMute",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRedVideoBlackLevel(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRedVideoBlackLevel action.

            :returns: "CurrentRedVideoBlackLevel"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetRedVideoBlackLevel", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentRedVideoBlackLevel",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRedVideoGain(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRedVideoGain action.

            :returns: "CurrentRedVideoGain"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetRedVideoGain", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentRedVideoGain",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSharpness(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSharpness action.

            :returns: "CurrentSharpness"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetSharpness", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentSharpness",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetStateVariables(self, InstanceID, StateVariableList, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetStateVariables action.

            :returns: "StateVariableValuePairs"
        """
        arguments = {
            "InstanceID": InstanceID,
            "StateVariableList": StateVariableList,
        }

        out_params = self.call_action("GetStateVariables", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableValuePairs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTransforms(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTransforms action.

            :returns: "CurrentTransformValues"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetTransforms", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTransformValues",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetVerticalKeystone(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetVerticalKeystone action.

            :returns: "CurrentVerticalKeystone"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetVerticalKeystone", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentVerticalKeystone",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetVolume(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetVolume action.

            :returns: "CurrentVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetVolume", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetVolumeDB(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetVolumeDB action.

            :returns: "CurrentVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetVolumeDB", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetVolumeDBRange(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetVolumeDBRange action.

            :returns: "MinValue", "MaxValue"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetVolumeDBRange", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MinValue", "MaxValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ListPresets(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ListPresets action.

            :returns: "CurrentPresetNameList"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("ListPresets", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentPresetNameList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SelectPreset(self, InstanceID, PresetName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SelectPreset action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "PresetName": PresetName,
        }

        self.call_action("SelectPreset", arguments=arguments, aspects=aspects)

        return

    def action_SetBlueVideoBlackLevel(self, InstanceID, DesiredBlueVideoBlackLevel, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBlueVideoBlackLevel action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredBlueVideoBlackLevel": DesiredBlueVideoBlackLevel,
        }

        self.call_action("SetBlueVideoBlackLevel", arguments=arguments, aspects=aspects)

        return

    def action_SetBlueVideoGain(self, InstanceID, DesiredBlueVideoGain, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBlueVideoGain action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredBlueVideoGain": DesiredBlueVideoGain,
        }

        self.call_action("SetBlueVideoGain", arguments=arguments, aspects=aspects)

        return

    def action_SetBrightness(self, InstanceID, DesiredBrightness, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBrightness action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredBrightness": DesiredBrightness,
        }

        self.call_action("SetBrightness", arguments=arguments, aspects=aspects)

        return

    def action_SetColorTemperature(self, InstanceID, DesiredColorTemperature, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetColorTemperature action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredColorTemperature": DesiredColorTemperature,
        }

        self.call_action("SetColorTemperature", arguments=arguments, aspects=aspects)

        return

    def action_SetContrast(self, InstanceID, DesiredContrast, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetContrast action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredContrast": DesiredContrast,
        }

        self.call_action("SetContrast", arguments=arguments, aspects=aspects)

        return

    def action_SetDefaultTransforms(self, DesiredDefaultTransformSettings, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDefaultTransforms action.
        """
        arguments = {
            "DesiredDefaultTransformSettings": DesiredDefaultTransformSettings,
        }

        self.call_action("SetDefaultTransforms", arguments=arguments, aspects=aspects)

        return

    def action_SetGreenVideoBlackLevel(self, InstanceID, DesiredGreenVideoBlackLevel, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetGreenVideoBlackLevel action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredGreenVideoBlackLevel": DesiredGreenVideoBlackLevel,
        }

        self.call_action("SetGreenVideoBlackLevel", arguments=arguments, aspects=aspects)

        return

    def action_SetGreenVideoGain(self, InstanceID, DesiredGreenVideoGain, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetGreenVideoGain action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredGreenVideoGain": DesiredGreenVideoGain,
        }

        self.call_action("SetGreenVideoGain", arguments=arguments, aspects=aspects)

        return

    def action_SetHorizontalKeystone(self, InstanceID, DesiredHorizontalKeystone, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetHorizontalKeystone action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredHorizontalKeystone": DesiredHorizontalKeystone,
        }

        self.call_action("SetHorizontalKeystone", arguments=arguments, aspects=aspects)

        return

    def action_SetLoudness(self, InstanceID, Channel, DesiredLoudness, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetLoudness action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredLoudness": DesiredLoudness,
        }

        self.call_action("SetLoudness", arguments=arguments, aspects=aspects)

        return

    def action_SetMute(self, InstanceID, Channel, DesiredMute, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetMute action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredMute": DesiredMute,
        }

        self.call_action("SetMute", arguments=arguments, aspects=aspects)

        return

    def action_SetRedVideoBlackLevel(self, InstanceID, DesiredRedVideoBlackLevel, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRedVideoBlackLevel action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredRedVideoBlackLevel": DesiredRedVideoBlackLevel,
        }

        self.call_action("SetRedVideoBlackLevel", arguments=arguments, aspects=aspects)

        return

    def action_SetRedVideoGain(self, InstanceID, DesiredRedVideoGain, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRedVideoGain action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredRedVideoGain": DesiredRedVideoGain,
        }

        self.call_action("SetRedVideoGain", arguments=arguments, aspects=aspects)

        return

    def action_SetSharpness(self, InstanceID, DesiredSharpness, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetSharpness action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredSharpness": DesiredSharpness,
        }

        self.call_action("SetSharpness", arguments=arguments, aspects=aspects)

        return

    def action_SetStateVariables(self, InstanceID, RenderingControlUDN, ServiceType, ServiceId, StateVariableValuePairs, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetStateVariables action.

            :returns: "StateVariableList"
        """
        arguments = {
            "InstanceID": InstanceID,
            "RenderingControlUDN": RenderingControlUDN,
            "ServiceType": ServiceType,
            "ServiceId": ServiceId,
            "StateVariableValuePairs": StateVariableValuePairs,
        }

        out_params = self.call_action("SetStateVariables", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetTransforms(self, InstanceID, DesiredTransformValues, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetTransforms action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredTransformValues": DesiredTransformValues,
        }

        self.call_action("SetTransforms", arguments=arguments, aspects=aspects)

        return

    def action_SetVerticalKeystone(self, InstanceID, DesiredVerticalKeystone, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetVerticalKeystone action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredVerticalKeystone": DesiredVerticalKeystone,
        }

        self.call_action("SetVerticalKeystone", arguments=arguments, aspects=aspects)

        return

    def action_SetVolume(self, InstanceID, Channel, DesiredVolume, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetVolume action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredVolume": DesiredVolume,
        }

        self.call_action("SetVolume", arguments=arguments, aspects=aspects)

        return

    def action_SetVolumeDB(self, InstanceID, Channel, DesiredVolume, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetVolumeDB action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredVolume": DesiredVolume,
        }

        self.call_action("SetVolumeDB", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class PrintBasic1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:PrintBasic:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'PrintBasic1'

    SERVICE_DEFAULT_VARIABLES = {
        "ColorSupported": { "data_type": "boolean", "default": None, "allowed_list": None},
        "Copies": { "data_type": "i4", "default": "1", "allowed_list": None},
        "DataSink": { "data_type": "uri", "default": None, "allowed_list": None},
        "DeviceId": { "data_type": "string", "default": None, "allowed_list": None},
        "DocumentFormat": { "data_type": "string", "default": None, "allowed_list": "['unknown', 'application/vnd.pwg-xhtml-print']"},
        "JobId": { "data_type": "i4", "default": "0", "allowed_list": None},
        "JobName": { "data_type": "string", "default": None, "allowed_list": None},
        "JobOriginatingUserName": { "data_type": "string", "default": None, "allowed_list": None},
        "MediaSize": { "data_type": "string", "default": None, "allowed_list": "['device-setting']"},
        "MediaType": { "data_type": "string", "default": None, "allowed_list": "['device-setting']"},
        "NumberUp": { "data_type": "string", "default": "1", "allowed_list": "['1', 'device-setting']"},
        "OrientationRequested": { "data_type": "string", "default": "portrait", "allowed_list": "['portrait', 'device-setting']"},
        "PrintQuality": { "data_type": "string", "default": "normal", "allowed_list": "['normal', 'device-setting']"},
        "PrinterLocation": { "data_type": "string", "default": None, "allowed_list": None},
        "PrinterName": { "data_type": "string", "default": None, "allowed_list": None},
        "Sides": { "data_type": "string", "default": "one-sided", "allowed_list": "['one-sided', 'device-setting']"},
        "XHTMLImageSupported": { "data_type": "string", "default": "image/jpeg", "allowed_list": "['image/jpeg']"},
    }

    SERVICE_EVENT_VARIABLES = {
        "JobEndState": { "data_type": "string", "default": None, "allowed_list": None},
        "JobIdList": { "data_type": "string", "default": None, "allowed_list": None},
        "JobMediaSheetsCompleted": { "data_type": "i4", "default": None, "allowed_list": None},
        "PrinterState": { "data_type": "string", "default": "idle", "allowed_list": "['idle', 'processing', 'stopped']"},
        "PrinterStateReasons": { "data_type": "string", "default": "none", "allowed_list": "['none', 'attention-required', 'media-jam', 'paused', 'door-open', 'media-low', 'media-empty', 'output-area-almost-full', 'output-area-full', 'marker-supply-low', 'marker-supply-empty', 'marker-failure', 'media-change-request']"},
    }

    def action_CancelJob(self, JobId, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CancelJob action.
        """
        arguments = {
            "JobId": JobId,
        }

        self.call_action("CancelJob", arguments=arguments, aspects=aspects)

        return

    def action_CreateJob(self, JobName, JobOriginatingUserName, DocumentFormat, Copies, Sides, NumberUp, OrientationRequested, MediaSize, MediaType, PrintQuality, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CreateJob action.

            :returns: "JobId", "DataSink"
        """
        arguments = {
            "JobName": JobName,
            "JobOriginatingUserName": JobOriginatingUserName,
            "DocumentFormat": DocumentFormat,
            "Copies": Copies,
            "Sides": Sides,
            "NumberUp": NumberUp,
            "OrientationRequested": OrientationRequested,
            "MediaSize": MediaSize,
            "MediaType": MediaType,
            "PrintQuality": PrintQuality,
        }

        out_params = self.call_action("CreateJob", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("JobId", "DataSink",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetJobAttributes(self, JobId, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetJobAttributes action.

            :returns: "JobName", "JobOriginatingUserName", "JobMediaSheetsCompleted"
        """
        arguments = {
            "JobId": JobId,
        }

        out_params = self.call_action("GetJobAttributes", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("JobName", "JobOriginatingUserName", "JobMediaSheetsCompleted",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPrinterAttributes(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPrinterAttributes action.

            :returns: "PrinterState", "PrinterStateReasons", "JobIdList", "JobId"
        """
        arguments = { }

        out_params = self.call_action("GetPrinterAttributes", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PrinterState", "PrinterStateReasons", "JobIdList", "JobId",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import traceback

def format_exc_lines():
    """
        Gets a 'format_exc' result and splits it into mutliple lines.
    """
    rtn_lines = traceback.format_exc().splitlines()
    return rtn_lines


"""
.. module:: x509
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains the classes and methods for working with X509 certificates.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from socket import gethostname
from typing import NamedTuple, Optional

from datetime import datetime

from OpenSSL import crypto


X509_DATETIME_FORMAT = '%Y%m%d%H%M%S%z'


class X509ExtensionInfo(NamedTuple):

    short_name: str
    data: str


class X509NameInfo(NamedTuple):

    country: str
    state_or_province: str
    locality: str
    organization: str
    unit: str
    common_name: str
    email_address: str


class X509CertificateInfo:

    def __init__(self, cert_obj: crypto.X509):
        self._cert_obj = cert_obj
        return

    @property
    def extensions(self):

        extdict = {}

        for ei in range(0, self._cert_obj.get_extension_count()):
            ext_obj = self._cert_obj.get_extension(ei)
            ex_short_name = ext_obj.get_short_name()
            ex_data = str(ext_obj)
            extdict[ex_short_name] = X509ExtensionInfo(ex_short_name, ex_data)

        return extdict

    @property
    def has_expired(self) -> bool:
        return self._cert_obj.has_expired()

    @property
    def issuer(self) -> X509NameInfo:
        iobj = self._cert_obj.get_issuer()
        iinfo = X509NameInfo(iobj.countryName, iobj.stateOrProvinceName, iobj.localityName, iobj.organizationName,
            iobj.organizationalUnitName, iobj.commonName, iobj.emailAddress)
        return iinfo
    
    @property
    def not_after(self) -> datetime:
        rtnval = self._cert_obj.get_notAfter()
        rtnval = datetime.strptime(rtnval, X509_DATETIME_FORMAT)
        return rtnval
    
    @property
    def not_before(self) -> datetime:
        rtnval = self._cert_obj.get_notBefore()
        rtnval = datetime.strptime(rtnval, X509_DATETIME_FORMAT)
        return rtnval

    @property
    def serial_no(self) -> int:
        rtnval = self._cert_obj.get_serial_number()
        return rtnval

    @property
    def sha256_digest(self) -> bytes:
        rtnval = self._cert_obj.digest("sha256")
        return rtnval

    @property
    def signature_algorithm(self) -> str:
        rtnval = self._cert_obj.get_signature_algorithm()
        return rtnval

    @property
    def subject(self) -> X509NameInfo:
        sobj = self._cert_obj.get_subject()
        sinfo = X509NameInfo(sobj.countryName, sobj.stateOrProvinceName, sobj.localityName, sobj.organizationName,
            sobj.organizationalUnitName, sobj.commonName, sobj.emailAddress)
        return sinfo

    @property
    def version(self) -> int:
        rtnval = self._cert_obj.get_version()
        return rtnval


def generate_self_signed_certificate(serial_no=1000, key_size=1024,
        country: Optional[str] = None, state_or_province: Optional[str] = None,
        locality: Optional[str] = None, organization: Optional[str] = None,
        unit: Optional[str] = None, common_name: Optional[str] = None,
        before_adjust_secs: Optional[int] = None, after_adjust_secs: Optional[int] = None):
    
    if country is None:
        country = "United States"
    if state_or_province is None:
        state_or_province = "Jefferson"
    if locality is None:
        locality = "Happy Camp"
    if organization is None:
        organization = "Automation Mojo"
    if unit is None:
        unit = "Automation Kit"
    if common_name is None:
        common_name = gethostname()

    # Create a key pair
    kp = crypto.PKey()
    kp.generate_key(crypto.TYPE_RSA, key_size)

    # Ceate a self-signed certificate
    cert = crypto.X509()

    subject = cert.get_subject()
    subject.C = country
    subject.ST = state_or_province
    subject.L = locality
    subject.O = organization
    subject.OU = unit
    subject.CN = common_name

    cert.set_serial_number(serial_no)

    if before_adjust_secs is None:
        before_adjust_secs = 0
    cert.gmtime_adj_notBefore(before_adjust_secs) # now

    if after_adjust_secs is None:
        years, days, hours, min, sec = 10, 365, 24, 60, 60
        after_adjust_secs = years * days * hours * min * sec
    cert.gmtime_adj_notAfter(after_adjust_secs)

    cert.set_issuer(subject)
    cert.set_pubkey(kp)

    cert.sign(kp, "sha1")

    # Create a P12 container
    p12 = crypto.PKCS12()
    p12.set_certificate(cert)

    cert_data = p12.export()

    return cert_data


def generate_signed_certificate(sig_key: crypto.PKey, issuer: crypto.X509Name, serial_no = 1000,
        key_size = 1024, country: Optional[str] = None, state_or_province: Optional[str] = None,
        locality: Optional[str] = None, organization: Optional[str] = None,
        unit: Optional[str] = None, common_name: Optional[str] = None,
        before_adjust_secs: Optional[int] = None, after_adjust_secs: Optional[int] = None):
    
    if country is None:
        country = "United States"
    if state_or_province is None:
        state_or_province = "Jefferson"
    if locality is None:
        locality = "Happy Camp"
    if organization is None:
        organization = "Automation Mojo"
    if unit is None:
        unit = "Automation Kit"
    if common_name is None:
        common_name = gethostname()

    # Create a key pair
    kp = crypto.PKey()
    kp.generate_key(crypto.TYPE_RSA, key_size)

    # Ceate a self-signed certificate
    cert = crypto.X509()

    subject = cert.get_subject()
    subject.C = country
    subject.ST = state_or_province
    subject.L = locality
    subject.O = organization
    subject.OU = unit
    subject.CN = common_name

    cert.set_serial_number(serial_no)

    if before_adjust_secs is None:
        before_adjust_secs = 0
    cert.gmtime_adj_notBefore(before_adjust_secs) # now

    if after_adjust_secs is None:
        years, days, hours, min, sec = 10, 365, 24, 60, 60
        after_adjust_secs = years * days * hours * min * sec
    cert.gmtime_adj_notAfter(after_adjust_secs)

    cert.set_pubkey(kp)

    cert.set_issuer(issuer)
    cert.sign(sig_key, "sha1")

    # Create a P12 container
    p12 = crypto.PKCS12()
    p12.set_certificate(cert)

    cert_data = p12.export()

    return cert_data


"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class CloudTransport1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:CloudTransport:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'CloudTransport1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {}

    def action_ConnectMethod(self, Host, MethodLine, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ConnectMethod action.

            :returns: "Identifier", "ConnectWriteCout"
        """
        arguments = {
            "Host": Host,
            "MethodLine": MethodLine,
        }

        out_params = self.call_action("ConnectMethod", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Identifier", "ConnectWriteCout",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_HTTPReadBody(self, Identifier, Size, CRLFFlag, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the HTTPReadBody action.

            :returns: "Body", "ReadLength", "ConnectReadCount"
        """
        arguments = {
            "Identifier": Identifier,
            "Size": Size,
            "CRLFFlag": CRLFFlag,
        }

        out_params = self.call_action("HTTPReadBody", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Body", "ReadLength", "ConnectReadCount",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_HTTPReadHeaders(self, Identifier, CRLFFlag, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the HTTPReadHeaders action.

            :returns: "Headers", "ReadLength", "ConnectReadCount"
        """
        arguments = {
            "Identifier": Identifier,
            "CRLFFlag": CRLFFlag,
        }

        out_params = self.call_action("HTTPReadHeaders", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Headers", "ReadLength", "ConnectReadCount",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_HTTPWriteBody(self, Identifier, Body, Size, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the HTTPWriteBody action.

            :returns: "ConnectWriteCount"
        """
        arguments = {
            "Identifier": Identifier,
            "Body": Body,
            "Size": Size,
        }

        out_params = self.call_action("HTTPWriteBody", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ConnectWriteCount",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_HTTPWriteHeaders(self, Identifier, Headers, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the HTTPWriteHeaders action.

            :returns: "ConnectWriteCount"
        """
        arguments = {
            "Identifier": Identifier,
            "Headers": Headers,
        }

        out_params = self.call_action("HTTPWriteHeaders", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ConnectWriteCount",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



from typing import List

from akit.interop.dns.dnsconst import DnsRecordClass, DnsRecordType

from akit.interop.dns.dnsquestion import DnsQuestion

from akit.interop.dns.dnsserver import DnsServer

def dns_query_name(name: str, rtype: DnsRecordType, rclass: DnsRecordClass):

    question = DnsQuestion(name, rtype, rclass)

    query_str = question.as_dns_string()

    return query_str

def dns_query_services(service_types: List[str], rtype: DnsRecordType=DnsRecordType.PTR, rclass: DnsRecordClass=DnsRecordClass.IN):
    
    service = DnsServer()
    
    service.start()

    service.wait()

    return

if __name__ == "__main__":
    from akit. interop.dns.dnsconst import DnsKnownServiceTypes

    server = dns_query_services([DnsKnownServiceTypes.SONOS])


"""
.. module:: runcommand
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A module that provides the RunCommand task class which implements
               the execution of a singular command based tasks in a workpacket.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

import os
import stat
import subprocess

from akit.paths import get_temporary_file
from akit.xformatting import indent_lines

from akit.workflow.tasks.taskbase import TaskBase

class RunCommand(TaskBase):

    def __init__(self, ordinal: str, label: str, task_info: dict, logger):
        super(RunCommand, self).__init__(ordinal, label, task_info, logger)
        self._command = task_info["command"]
        return

    @property
    def command(self) -> str:
        return self._command

    def execute(self, parameters: Optional[dict]=None, topology: Optional[dict]=None, **kwargs) -> int:

        self._logger.info("STEP: %s - %s" % (self._label, self._ordinal))

        self._logger.info("Running Command: %s" % self._command)

        proc = subprocess.Popen([self._command], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = proc.communicate()
        exit_code = proc.wait()

        stdout = stdout.decode()
        stderr = stderr.decode()

        log_msg_lines = [
            "RESULT CODE: %d" % exit_code,
            "STDOUT:",
            indent_lines(stdout, level=1),
            "STDERR:",
            indent_lines(stderr, level=1),
        ]

        log_msg = os.linesep.join(log_msg_lines)
        if exit_code == 0:
            self._logger.info(log_msg)
        else:
            self._logger.error(log_msg)

        return exit_code


"""
.. module:: soap
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains classes and constants for working with SOAP message processing.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import enum
import re

from xml.etree.ElementTree import Element, SubElement, QName, ParseError
from xml.etree.ElementTree import tostring as xml_tostring
from xml.etree.ElementTree import fromstring as xml_fromstring
from xml.etree.ElementTree import register_namespace

from akit.compat import str_cast

from akit.exceptions import AKitCommunicationsProtocolError
from akit.interop.upnp.upnperrors import UPNP_ERROR_TEST_LOOKUP

NS_SOAP_ENV = "http://schemas.xmlsoap.org/soap/envelope/"
NS_SOAP_ENC = "http://schemas.xmlsoap.org/soap/encoding/"
NS_XSI = "http://www.w3.org/1999/XMLSchema-instance"
NS_XSD = "http://www.w3.org/1999/XMLSchema"

NS_UPNP_CONTROL = "urn:schemas-upnp-org:control-1-0"
NS_UPNP_EVENT = "urn:schemas-upnp-org:event-1-0"

URI_SOAP_ENCODING = "http://schemas.xmlsoap.org/soap/encoding/"

XML_DOCUMENT_DECLARATION = '<?xml version="1.0" encoding="utf-8"?>'

SOAP_TIMEOUT = 60

PYTHON_TO_SOAP_TYPE_MAP = {
    bytes: 'xsd:string',
    str: 'xsd:string',
    int: 'xsd:int',
    float: 'xsd:float',
    bool: 'xsd:boolean'
}

class SOAPError(AKitCommunicationsProtocolError):
    """
        Error that is raised when a Soap error occurs.
    """


class SOAPProtocolError(AKitCommunicationsProtocolError):
    """
        Error that is raised when a Soap protocol error occurs.
    """

def remove_extraneous_xml_declarations(xml_str):
    """
        Special method for correcting XML content that has an extra XML declaration.
    """
    xml_declaration = ''
    if xml_str.startswith('<?xml'):
        xml_declaration, xml_str = xml_str.split('?>', maxsplit=1)
    xml_declaration += '?>'
    xml_str = re.sub(r'<\?xml.*?\?>', '', xml_str, flags=re.I)
    return xml_declaration + xml_str

class SoapProcessor:
    """
        The Soap processor object stores encoding and decoding settings for processing the Soap messages
        senti to and from a UPNP device.
    """
    def __init__(self, encoding=URI_SOAP_ENCODING, envelope_attrib=None, typed=None):
        self._encoding = encoding
        self._envelope_attrib = envelope_attrib
        self._typed = typed
        return

    def create_request(self, action_name: str, arguments: dict, encoding=None, envelope_attrib=None, typed=None):
        """
            Creates a Soap request for a call on the specified action and with the specified arguments.
        """
        register_namespace('', None)

        if encoding is None:
            encoding = self._encoding
        if envelope_attrib is None:
            envelope_attrib = self._envelope_attrib
        if typed is None:
            typed = self._typed

        envelope = Element("s:Envelope")
        if envelope_attrib:
            for eakey, eaval in envelope_attrib:
                envelope.attrib.update({eakey: eaval})
        else:
            envelope.attrib.update({'xmlns:s': NS_SOAP_ENV})
            envelope.attrib.update({'s:encodingStyle': encoding})

        body = SubElement(envelope, "s:Body")

        if typed:
            methElement = SubElement(body, "u:" + action_name)
            methElement.attrib.update({'xmlns:u': typed})
        else:
            methElement = SubElement(body, action_name)

        if arguments:
            for arg_name, arg_val in arguments.items():
                if isinstance(arg_val, enum.Enum):
                    arg_val = arg_val.value

                py_type = type(arg_val)
                soap_type = PYTHON_TO_SOAP_TYPE_MAP[py_type]

                if soap_type == 'xsd:string':
                    arg_val = arg_val # pylint: disable=self-assigning-variable
                elif soap_type == 'xsd:int' or soap_type == 'xsd:float':
                    arg_val = str(arg_val)
                elif soap_type == 'xsd:boolean':
                    arg_val = "1" if arg_val else "0"

                argElement = SubElement(methElement, arg_name)
                #if typed and soap_type:
                #    if not isinstance(type, QName):
                #        soap_type = QName(NS_XSD, soap_type)
                #    argElement.set(NS_XSI + "type", soap_type)

                argElement.text = arg_val
        else:
            methElement.text = ""

        envelope_content = xml_tostring(envelope, short_empty_elements=False)
        content = XML_DOCUMENT_DECLARATION + str_cast(envelope_content)

        return content

    def create_response(self, action_name: str, arguments: dict, encoding=None, envelope_attrib=None, typed=None):
        """
            Creates a Soap response to the action with the specified arguments.
        """
        register_namespace('', None)

        if encoding is None:
            encoding = self._encoding
        if envelope_attrib is None:
            envelope_attrib = self._envelope_attrib
        if typed is None:
            typed = self._typed

        envelope = Element("s:Envelope")
        if envelope_attrib:
            for eakey, eaval in envelope_attrib:
                envelope.attrib.update({eakey: eaval})
        else:
            envelope.attrib.update({'xmlns:s': NS_SOAP_ENV})
            envelope.attrib.update({'s:encodingStyle': encoding})

        body = SubElement(envelope, "s:Body")

        action_name_tag_name = action_name + "Response"
        methElement = SubElement(body, action_name_tag_name)
        if encoding:
            methElement.set(NS_SOAP_ENV + "encodingStyle", encoding)

        if arguments:
            for arg_name, arg_val in arguments.items():
                py_type = type(arg_val)
                soap_type = PYTHON_TO_SOAP_TYPE_MAP[py_type]

                if soap_type == 'xsd:string':
                    arg_val = arg_val # pylint: disable=self-assigning-variable
                elif soap_type == 'xsd:int' or soap_type == 'xsd:float':
                    arg_val = str(arg_val)
                elif soap_type == 'xsd:boolean':
                    arg_val = "1" if arg_val else "0"

                argElement = SubElement(methElement, arg_name)
                if typed and soap_type:
                    if not isinstance(type, QName):
                        arg_type = QName(NS_XSD, soap_type)
                    argElement.set(NS_XSI + "type", soap_type)

                argElement.text = arg_val
        else:
            methElement.text = ""

        envelope_content = xml_tostring(envelope, short_empty_elements=False)
        content = XML_DOCUMENT_DECLARATION + "\n" + str_cast(envelope_content)

        return content

    def parse_response(self, action_name, content, encoding=None, envelope_attrib=None, typed=None):
        """
            Parses a response from the server with the given action name and content.
        """
        register_namespace('', None)

        if encoding is None:
            encoding = self._encoding
        if envelope_attrib is None:
            envelope_attrib = self._envelope_attrib
        if typed is None:
            typed = self._typed

        try:
            docNode = xml_fromstring(content)
        except ParseError:
            # Try removing any extra XML declarations in case there are more than one.
            # This sometimes happens when a device sends its own XML config files.
            content = remove_extraneous_xml_declarations(content)
            docNode = xml_fromstring(content)
        except ValueError:
            # This can occur when requests returns a `str` (unicode) but there's also an XML
            # declaration, which lxml doesn't like.
            docNode = xml_fromstring(content.encode('utf8'))

        resp_body = None
        if typed:
            resp_body = docNode.find(".//{%s}%sResponse" % (typed, action_name))
        else:
            resp_body = docNode.find(".//%sResponse" % action_name)

        if resp_body is None:
            msg = ('Returned XML did not include an element which matches namespace %r and tag name'
                   ' \'%sResponse\'.' % (typed, action_name))
            print(msg + '\n' + xml_tostring(docNode, short_empty_elements=False).decode('utf8'))
            raise SOAPProtocolError(msg)

        # Sometimes devices return XML strings as their argument values without escaping them with
        # CDATA. This checks to see if the argument has been parsed as XML and un-parses it if so.
        resp_dict = {}
        for arg in resp_body.getchildren():
            children = arg.getchildren()
            if children:
                resp_dict[arg.tag] = "\n".join(xml_tostring(x) for x in children)
            else:
                if arg.text is None:
                    resp_dict[arg.tag] = ""
                else:
                    resp_dict[arg.tag] = arg.text

        return resp_dict

    def parse_response_error_for_upnp(self, action_name, content, status_code, extra=None, encoding=None, envelope_attrib=None, typed=None):
        """
            Parse response error for a upnp response.
        """
        register_namespace('', None)

        if encoding is None:
            encoding = self._encoding
        if envelope_attrib is None:
            envelope_attrib = self._envelope_attrib
        if typed is None:
            typed = self._typed

        try:
            docNode = xml_fromstring(content)
        except ParseError:
            # Try removing any extra XML declarations in case there are more than one.
            # This sometimes happens when a device sends its own XML config files.
            content = remove_extraneous_xml_declarations(content)
            docNode = xml_fromstring(content)
        except ValueError:
            # This can occur when requests returns a `str` (unicode) but there's also an XML
            # declaration, which lxml doesn't like.
            docNode = xml_fromstring(content.encode('utf8'))

        resp_body = None
        if typed:
            resp_body = docNode.find(".//{%s}Fault" % (NS_SOAP_ENV,))
        else:
            resp_body = docNode.find(".//Fault")

        if resp_body is None:
            msg = ('Returned XML did not include an element which matches namespace %r and tag name'
                   ' \'%sFault\'.' % (typed, action_name))
            print(msg + '\n' + xml_tostring(docNode, short_empty_elements=False).decode('utf8'))
            raise SOAPProtocolError(msg)

        # Lets try to extract the XML response error information
        try:
            faultCode = resp_body.find(".//faultcode").text
            faultString = resp_body.find(".//faultstring").text
            detail = resp_body.find(".//detail")
            upnpErrorNode = detail.find(".//{%s}UPnPError" % NS_UPNP_CONTROL)
            errorCode = int(upnpErrorNode.find(".//{%s}errorCode" % NS_UPNP_CONTROL).text)
            errorDescription = upnpErrorNode.find(".//{%s}errorDescription" % NS_UPNP_CONTROL)
            if errorDescription is None:
                errorDescription = UPNP_ERROR_TEST_LOOKUP.get(errorCode, "Unknown error.")

        except Exception as xcpt:
            errmsg = "Unable to process xml response: status=%r\n%s" % (status_code, content)
            if extra is not None:
                errmsg += "EXTRA:\n%s" % extra
            raise SOAPProtocolError(errmsg) from xcpt

        return errorCode, errorDescription



"""
.. module:: paths
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains the :class:`TaskBase` object which is used as the base.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

# pylint: disable=global-statement

from cgitb import lookup
from typing import List, Optional

import os
import tempfile

from akit.environment.context import Context
from akit.environment.contextpaths import ContextPaths
from akit.environment.variables import AKIT_VARIABLES

from akit.exceptions import AKitRuntimeError

DIR_DIAGNOSTICS_DIRECTORY = None
DIR_RESULTS_DIRECTORY = None
DIR_TESTRESULTS_DIRECTORY = None

TRANSLATE_TABLE_NORMALIZE_FOR_PATH = str.maketrans(",.:;", "    ")

DEFAULT_PATH_EXPANSIONS = [
    os.path.expanduser,
    os.path.expandvars,
    os.path.abspath
]

def expand_path(path_in, expansions=DEFAULT_PATH_EXPANSIONS):

    path_out = path_in
    for expansion_func in expansions:
        path_out = expansion_func(path_out)

    return path_out

def collect_python_modules(search_dir: str, max_depth=None) -> List[str]:
    """
        Walks a directory tree of python modules and collects the names
        of all of the python module files or .py files.  This method allows
        for python namespaces by not forcing the root folder to contain a
        __init__.py file.

        :params searchdir: The root directory to search when collecting python modules.
    """
    pyfiles = []

    search_dir = os.path.abspath(search_dir)
    search_dir_len = len(search_dir)

    for root, _, files in os.walk(search_dir, topdown=True):

        if max_depth is not None:
            dir_leaf = root[search_dir_len:].strip(os.path.sep)
            depth = 0
            if dir_leaf != '':
                dir_leaf_parts = dir_leaf.split(os.path.sep)
                depth = len(dir_leaf_parts)
                if depth > max_depth:
                    break

        for fname in files:
            fbase, fext = os.path.splitext(fname)
            if fext == '.py' and fbase != "__init__":
                ffull = os.path.join(root, fname)
                pyfiles.append(ffull)

    return pyfiles

def ensure_directory_is_package(package_dir: str, package_title: Optional[str] = None):
    """
        Ensures that a directory is represented to python as a package by checking to see if the
        directory has an __init__.py file and if not it adds one.

        :param package_dir: The direcotry to represent as a package.
        :param package_title: Optional title to be written into the documentation string in the package file.
    """
    package_dir_init = os.path.join(package_dir, "__init__.py")
    if not os.path.exists(package_dir_init):
        with open(package_dir_init, 'w') as initf:
            initf.write('"""\n')
            if package_title is not None:
                initf.write('   %s\n' % package_title)
            initf.write('"""\n')
    return

def get_directory_for_code_container(container: str) -> str:
    """
        Returns the directory for a code container (module or package)

        :param container: The code container you want to get a directory for.

        :returns: The string that represents the parent directory of the code
                  container specified.
    """
    if hasattr(container, '__path__'):
        container_dir = str(container.__path__[0]).rstrip(os.sep)
    elif hasattr(container, '__file__'):
        container_dir = os.path.dirname(container.__file__).rstrip(os.sep)
    else:
        raise AKitRuntimeError("Unable to get parent dir for module") from None

    return container_dir

def get_expanded_path(path: str) -> str:
    """
        Returns a path expanded using expanduser, expandvars and abspath for
        the provided path.

        :param path: A path which you want to expand to a full path, expanding the
                     user, variables and relative path syntax.

        :returns: The expanded path
    """
    exp_path = os.path.abspath(os.path.expandvars(os.path.expanduser(path)))
    return exp_path

def get_filename_for_credentials() -> str:
    """
        Returns the path to the credentials file.
    """
    filename = None

    ctx = Context()
    file_confg = ctx.lookup(ContextPaths.CONFIG_FILE_CREDENTIALS)
    filename = get_expanded_path(file_confg)

    return filename

def get_filename_for_landscape() -> str:
    """
        Returns the path to the landscape file.
    """
    filename = None

    ctx = Context()
    file_confg = ctx.lookup(ContextPaths.CONFIG_FILE_LANDSCAPE)
    filename = get_expanded_path(file_confg)

    return filename

def get_filename_for_runtime() -> str:
    """
        Returns the path to the landscape file.
    """
    filename = None

    ctx = Context()
    file_confg = ctx.lookup(ContextPaths.CONFIG_FILE_RUNTIME)
    filename = get_expanded_path(file_confg)

    return filename

def get_filename_for_topology() -> str:
    """
        Returns the path to the topology file.
    """
    filename = None

    ctx = Context()
    filename = get_expanded_path(ctx.lookup(ContextPaths.CONFIG_FILE_TOPOLOGY))

    return filename

def get_path_for_artifacts(label: str) -> str:
    """
        Returns a path in the form (testresultdir)/artifacts/(label)

        :param label: A label to associate with the collection of artifacts. The label is used for
                      the name of the artifact container folder.

        :returns: A path that is descendant from (testresultdir)/artifacts
    """
    trdir = get_path_for_output()
    afdir = os.path.join(trdir, "artifacts", label)

    if not os.path.exists(afdir):
        os.makedirs(afdir)

    return afdir

def get_path_for_output(create=True) -> str:
    """
        Returns the timestamped path where test results and artifacts are deposited to
    """
    global DIR_RESULTS_DIRECTORY

    if DIR_RESULTS_DIRECTORY is None:
        ctx = Context()

        DIR_RESULTS_DIRECTORY = get_expanded_path(ctx.lookup(ContextPaths.OUTPUT_DIRECTORY))
        if create and not os.path.exists(DIR_RESULTS_DIRECTORY):
            os.makedirs(DIR_RESULTS_DIRECTORY)

    return DIR_RESULTS_DIRECTORY

def get_path_for_diagnostics(label: str) -> str:
    """
        Returns a path in the form (testresultdir)/diagnostics/(label)

        :param label: A label to associate with the collection of diagnostic captures.
                      The label is used for the name of the diagnostic container folder.

        :returns: A path that is descendant from (testresultdir)/diagnostics

        .. notes: Paths that are relative to the output directory should always call
                  get_path_for_output instead of calling for the directory of any
                  one job type.
    """

    
    trdir = get_path_for_output()
    diagnostics_dir = os.path.join(trdir, "diagnostics", label)

    if not os.path.exists(diagnostics_dir):
        os.makedirs(diagnostics_dir)

    return diagnostics_dir

def get_path_for_testresults() -> str:
    """
        Returns the path to the testresults directory.

        :returns: The path to the testresults directory

        ..note : This path is generally only used by the commandline because
        it can intentionally provide a different result than get_path_for_output
        if the current job is not a TESTRUN job.  If your running in a TESTRUN job,
        utilize get_path_for_output instead of this API.
    """

    global DIR_TESTRESULTS_DIRECTORY

    if DIR_TESTRESULTS_DIRECTORY is None:
        ctx = Context()
        configuration = ctx.lookup("/configuration")

        tr_dir = configuration.lookup("/paths/testresults")

        DIR_TESTRESULTS_DIRECTORY = tr_dir

        if not os.path.exists(DIR_TESTRESULTS_DIRECTORY):
            os.makedirs(DIR_TESTRESULTS_DIRECTORY)

    return DIR_TESTRESULTS_DIRECTORY

def get_summary_html_template_source() -> str:
    """
        Looks up a source path for the summary html template.

        :returns: The path to the html summary template
    """
    
    from akit.environment.context import Context

    ctx = Context()

    template = ctx.lookup(ContextPaths.FILE_RESULTS_TEMPLATE)
    
    if template is None:
        errmsg = "Error attempting to lookup the summary html template source."
        raise AKitRuntimeError(errmsg)

    return template

def get_summary_static_resource_dest_dir(create=True) -> str:
    """
        Returns the path where the static resources for test summaries should be published.
    """
    ctx = Context()
    res_dir = ctx.lookup(ContextPaths.DIR_RESULTS_RESOURCE_DEST)

    if create and not os.path.exists(res_dir):
        os.makedirs(res_dir)

    return res_dir

def get_summary_static_resource_src_dir() -> str:
    """
        Returns the path that is the source path for the test summary static resources.
    """
    ctx = Context()
    res_dir = ctx.lookup(ContextPaths.DIR_RESULTS_RESOURCE_SRC)

    return res_dir

def get_temporary_directory() -> str:
    """
        Returns the path of a temporary directory in the output directory.
    """
    temp_dir = os.path.join(get_path_for_output(), "temp")
    return temp_dir

def get_temporary_file(suffix: str = '', prefix: str = '') -> str:
    """
        Returns the path of a temporary file in the output directory.
    """
    tempdir = get_temporary_directory()

    tmpfile = tempfile.mktemp(suffix=suffix, prefix=prefix, dir=tempdir)

    return tmpfile

def normalize_name_for_path(name: str) -> str:
    """
        Normalizes a path string by replacing ",.:;" with space and then removing
        white space.

        :param name: A name as a str which is to be normalized to allow it to be used in a path.

        :returns: The normalized string which can be used in a path.
    """
    norm_name = name.translate(TRANSLATE_TABLE_NORMALIZE_FOR_PATH).replace(" ", "")
    return norm_name



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class AVTransport3ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:AVTransport:3' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'AVTransport3'

    SERVICE_DEFAULT_VARIABLES = {
        "AVTransportURI": { "data_type": "string", "default": None, "allowed_list": None},
        "AVTransportURIMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "AbsoluteCounterPosition": { "data_type": "ui4", "default": None, "allowed_list": None},
        "AbsoluteTimePosition": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentMediaCategory": { "data_type": "string", "default": None, "allowed_list": "['NO_MEDIA', 'TRACK_AWARE', 'TRACK_UNAWARE']"},
        "CurrentMediaDuration": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentPlayMode": { "data_type": "string", "default": "NORMAL", "allowed_list": "['NORMAL']"},
        "CurrentRecordQualityMode": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTrack": { "data_type": "ui4", "default": None, "allowed_list": None},
        "CurrentTrackDuration": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTrackMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTrackURI": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentTransportActions": { "data_type": "string", "default": None, "allowed_list": None},
        "DRMState": { "data_type": "string", "default": "UNKNOWN", "allowed_list": "['OK']"},
        "NextAVTransportURI": { "data_type": "string", "default": None, "allowed_list": None},
        "NextAVTransportURIMetaData": { "data_type": "string", "default": None, "allowed_list": None},
        "NumberOfTracks": { "data_type": "ui4", "default": None, "allowed_list": None},
        "PlaybackStorageMedium": { "data_type": "string", "default": None, "allowed_list": None},
        "PossiblePlaybackStorageMedia": { "data_type": "string", "default": None, "allowed_list": None},
        "PossibleRecordQualityModes": { "data_type": "string", "default": None, "allowed_list": None},
        "PossibleRecordStorageMedia": { "data_type": "string", "default": None, "allowed_list": None},
        "RecordMediumWriteStatus": { "data_type": "string", "default": None, "allowed_list": None},
        "RecordStorageMedium": { "data_type": "string", "default": None, "allowed_list": None},
        "RelativeCounterPosition": { "data_type": "i4", "default": None, "allowed_list": None},
        "RelativeTimePosition": { "data_type": "string", "default": None, "allowed_list": None},
        "SyncOffset": { "data_type": "string", "default": None, "allowed_list": None},
        "TransportPlaySpeed": { "data_type": "string", "default": None, "allowed_list": "['1']"},
        "TransportState": { "data_type": "string", "default": None, "allowed_list": "['STOPPED', 'PLAYING']"},
        "TransportStatus": { "data_type": "string", "default": None, "allowed_list": "['OK', 'ERROR_OCCURRED']"},
    }

    SERVICE_EVENT_VARIABLES = {
        "LastChange": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AdjustSyncOffset(self, InstanceID, Adjustment, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AdjustSyncOffset action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Adjustment": Adjustment,
        }

        self.call_action("AdjustSyncOffset", arguments=arguments, aspects=aspects)

        return

    def action_GetCurrentTransportActions(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCurrentTransportActions action.

            :returns: "Actions"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetCurrentTransportActions", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Actions",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDRMState(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDRMState action.

            :returns: "CurrentDRMState"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetDRMState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentDRMState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDeviceCapabilities(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDeviceCapabilities action.

            :returns: "PlayMedia", "RecMedia", "RecQualityModes"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetDeviceCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PlayMedia", "RecMedia", "RecQualityModes",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMediaInfo(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMediaInfo action.

            :returns: "NrTracks", "MediaDuration", "CurrentURI", "CurrentURIMetaData", "NextURI", "NextURIMetaData", "PlayMedium", "RecordMedium", "WriteStatus"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetMediaInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NrTracks", "MediaDuration", "CurrentURI", "CurrentURIMetaData", "NextURI", "NextURIMetaData", "PlayMedium", "RecordMedium", "WriteStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMediaInfo_Ext(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMediaInfo_Ext action.

            :returns: "CurrentType", "NrTracks", "MediaDuration", "CurrentURI", "CurrentURIMetaData", "NextURI", "NextURIMetaData", "PlayMedium", "RecordMedium", "WriteStatus"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetMediaInfo_Ext", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentType", "NrTracks", "MediaDuration", "CurrentURI", "CurrentURIMetaData", "NextURI", "NextURIMetaData", "PlayMedium", "RecordMedium", "WriteStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPlaylistInfo(self, InstanceID, PlaylistType, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPlaylistInfo action.

            :returns: "PlaylistInfo"
        """
        arguments = {
            "InstanceID": InstanceID,
            "PlaylistType": PlaylistType,
        }

        out_params = self.call_action("GetPlaylistInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PlaylistInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPositionInfo(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPositionInfo action.

            :returns: "Track", "TrackDuration", "TrackMetaData", "TrackURI", "RelTime", "AbsTime", "RelCount", "AbsCount"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetPositionInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Track", "TrackDuration", "TrackMetaData", "TrackURI", "RelTime", "AbsTime", "RelCount", "AbsCount",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetStateVariables(self, InstanceID, StateVariableList, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetStateVariables action.

            :returns: "StateVariableValuePairs"
        """
        arguments = {
            "InstanceID": InstanceID,
            "StateVariableList": StateVariableList,
        }

        out_params = self.call_action("GetStateVariables", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableValuePairs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSyncOffset(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSyncOffset action.

            :returns: "CurrentSyncOffset"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetSyncOffset", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentSyncOffset",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTransportInfo(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTransportInfo action.

            :returns: "CurrentTransportState", "CurrentTransportStatus", "CurrentSpeed"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetTransportInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTransportState", "CurrentTransportStatus", "CurrentSpeed",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTransportSettings(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTransportSettings action.

            :returns: "PlayMode", "RecQualityMode"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetTransportSettings", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PlayMode", "RecQualityMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Next(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Next action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Next", arguments=arguments, aspects=aspects)

        return

    def action_Pause(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Pause action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Pause", arguments=arguments, aspects=aspects)

        return

    def action_Play(self, InstanceID, Speed, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Play action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Speed": Speed,
        }

        self.call_action("Play", arguments=arguments, aspects=aspects)

        return

    def action_Previous(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Previous action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Previous", arguments=arguments, aspects=aspects)

        return

    def action_Record(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Record action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Record", arguments=arguments, aspects=aspects)

        return

    def action_Seek(self, InstanceID, Unit, Target, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Seek action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Unit": Unit,
            "Target": Target,
        }

        self.call_action("Seek", arguments=arguments, aspects=aspects)

        return

    def action_SetAVTransportURI(self, InstanceID, CurrentURI, CurrentURIMetaData, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAVTransportURI action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "CurrentURI": CurrentURI,
            "CurrentURIMetaData": CurrentURIMetaData,
        }

        self.call_action("SetAVTransportURI", arguments=arguments, aspects=aspects)

        return

    def action_SetNextAVTransportURI(self, InstanceID, NextURI, NextURIMetaData, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetNextAVTransportURI action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NextURI": NextURI,
            "NextURIMetaData": NextURIMetaData,
        }

        self.call_action("SetNextAVTransportURI", arguments=arguments, aspects=aspects)

        return

    def action_SetPlayMode(self, InstanceID, NewPlayMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetPlayMode action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NewPlayMode": NewPlayMode,
        }

        self.call_action("SetPlayMode", arguments=arguments, aspects=aspects)

        return

    def action_SetRecordQualityMode(self, InstanceID, NewRecordQualityMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRecordQualityMode action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NewRecordQualityMode": NewRecordQualityMode,
        }

        self.call_action("SetRecordQualityMode", arguments=arguments, aspects=aspects)

        return

    def action_SetStateVariables(self, InstanceID, AVTransportUDN, ServiceType, ServiceId, StateVariableValuePairs, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetStateVariables action.

            :returns: "StateVariableList"
        """
        arguments = {
            "InstanceID": InstanceID,
            "AVTransportUDN": AVTransportUDN,
            "ServiceType": ServiceType,
            "ServiceId": ServiceId,
            "StateVariableValuePairs": StateVariableValuePairs,
        }

        out_params = self.call_action("SetStateVariables", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetStaticPlaylist(self, InstanceID, PlaylistData, PlaylistDataLength, PlaylistOffset, PlaylistTotalLength, PlaylistMIMEType, PlaylistExtendedType, PlaylistStartObj, PlaylistStartGroup, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetStaticPlaylist action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "PlaylistData": PlaylistData,
            "PlaylistDataLength": PlaylistDataLength,
            "PlaylistOffset": PlaylistOffset,
            "PlaylistTotalLength": PlaylistTotalLength,
            "PlaylistMIMEType": PlaylistMIMEType,
            "PlaylistExtendedType": PlaylistExtendedType,
            "PlaylistStartObj": PlaylistStartObj,
            "PlaylistStartGroup": PlaylistStartGroup,
        }

        self.call_action("SetStaticPlaylist", arguments=arguments, aspects=aspects)

        return

    def action_SetStreamingPlaylist(self, InstanceID, PlaylistData, PlaylistDataLength, PlaylistMIMEType, PlaylistExtendedType, PlaylistStep, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetStreamingPlaylist action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "PlaylistData": PlaylistData,
            "PlaylistDataLength": PlaylistDataLength,
            "PlaylistMIMEType": PlaylistMIMEType,
            "PlaylistExtendedType": PlaylistExtendedType,
            "PlaylistStep": PlaylistStep,
        }

        self.call_action("SetStreamingPlaylist", arguments=arguments, aspects=aspects)

        return

    def action_SetSyncOffset(self, InstanceID, NewSyncOffset, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetSyncOffset action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "NewSyncOffset": NewSyncOffset,
        }

        self.call_action("SetSyncOffset", arguments=arguments, aspects=aspects)

        return

    def action_Stop(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Stop action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("Stop", arguments=arguments, aspects=aspects)

        return

    def action_SyncPause(self, InstanceID, PauseTime, ReferenceClockId, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SyncPause action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "PauseTime": PauseTime,
            "ReferenceClockId": ReferenceClockId,
        }

        self.call_action("SyncPause", arguments=arguments, aspects=aspects)

        return

    def action_SyncPlay(self, InstanceID, Speed, ReferencePositionUnits, ReferencePosition, ReferencePresentationTime, ReferenceClockId, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SyncPlay action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Speed": Speed,
            "ReferencePositionUnits": ReferencePositionUnits,
            "ReferencePosition": ReferencePosition,
            "ReferencePresentationTime": ReferencePresentationTime,
            "ReferenceClockId": ReferenceClockId,
        }

        self.call_action("SyncPlay", arguments=arguments, aspects=aspects)

        return

    def action_SyncStop(self, InstanceID, StopTime, ReferenceClockId, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SyncStop action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "StopTime": StopTime,
            "ReferenceClockId": ReferenceClockId,
        }

        self.call_action("SyncStop", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import random

from enum import IntEnum
from string import hexdigits


class WEP_BIT_COUNT(IntEnum):
    BITS_64 = 64
    BITS_128 = 128
    BITS_152 = 152
    BITS_256 = 256

WEP_BITS_TO_HEX_CHARS = {
    WEP_BIT_COUNT.BITS_64: 5,
    WEP_BIT_COUNT.BITS_128: 13,
    WEP_BIT_COUNT.BITS_152: 16,
    WEP_BIT_COUNT.BITS_256: 29
}


def create_random_wep_key(bit_count: WEP_BIT_COUNT=WEP_BIT_COUNT.BITS_128):
    """
    """
    digit_count = WEP_BITS_TO_HEX_CHARS[bit_count]
    key = random.sample(hexdigits, digit_count)
    return key




"""
.. module:: landscapedevice
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`LandscapeDevice` class.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Callable, FrozenSet, List, Optional, OrderedDict, Union, TYPE_CHECKING

import bisect
import os
import threading
import weakref

from enum import Enum
from datetime import datetime

from akit.exceptions import AKitRuntimeError, AKitSemanticError
from akit.friendlyidentifier import FriendlyIdentifier
from akit.interfaces.icommandrunner import ICommandRunner
from akit.xfeature import FeatureAttachedObject

from akit.xlogging.foundations import getAutomatonKitLogger

if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape
    from akit.interop.landscaping.landscapedeviceextension import LandscapeDeviceExtension
    from akit.interop.agents.sshagent import SshAgent
    from akit.interop.upnp.devices.upnprootdevice import UpnpRootDevice

class LandscapeDevice(FeatureAttachedObject):
    """
        The base class for all landscape devices.  The :class:`LandscapeDevice' represents attributes that are common
        to all connected devices and provides attachements points and mechanisms for adding DeviceExtentions to
        the :class:`LandscapeDevice` device.
    """

    # Base landscape devices don't have any feature tags, but we want all devices to have feature
    # tag capability so we can filter all devices based on feature tags, whether they have
    # tags or not.
    FEATURE_TAGS = []

    logger = getAutomatonKitLogger()

    def __init__(self, lscape: "Landscape", friendly_id: FriendlyIdentifier, device_type: str, device_config: dict):
        super().__init__()

        self._lscape_ref = weakref.ref(lscape)
        self._friendly_id = friendly_id
        self._device_type = device_type
        self._device_config = device_config

        self._device_lock = threading.RLock()

        self._contacted_first = None
        self._contacted_last = None
        self._is_watched = None
        self._is_isolated = None

        self._upnp: "UpnpRootDevice" = None
        self._ssh: "SshAgent" = None
        self._power = None
        self._serial = None

        self._preferred_command_interface = "ssh"

        self._match_functions = {}

        self._credentials = {}
        self._ssh_credentials = []

        if "credentials" in device_config:
            lscape_credentials = lscape.credentials
            for cred_key in device_config["credentials"]:
                if cred_key in lscape_credentials:
                    cred_info = lscape_credentials[cred_key]
                    self._credentials[cred_key] = cred_info
                    if cred_info.category == "ssh":
                        self._ssh_credentials.append(cred_info)

        return

    @property
    def contacted_first(self) -> datetime:
        """
            A datetime stamp of when this device was first contacted
        """
        return self._contacted_first

    @property
    def contacted_last(self) -> datetime:
        """
            A datetime stamp of when this device was last contacted
        """
        return self._contacted_last

    @property
    def device_config(self) -> dict:
        """
            A dictionary of the configuration information for this device.
        """
        return self._device_config

    @property
    def device_lock(self) -> threading.RLock:
        """
            Returns the lock for the device.
            
            ..note: Use with caution
        """
        return self._device_lock

    @property
    def device_type(self) -> str:
        """
            A string representing the type of device.
        """
        return self._device_type

    @property
    def friendly_id(self) -> FriendlyIdentifier:
        """
            The friendly identifier for this device, this is generally the identifier provided
            by the coordinator that created the device instance.
        """
        return self._friendly_id

    @property
    def has_ssh_credential(self) -> bool:
        """
            A boolean value indicating whether this device has an SSH credential.
        """
        has_creds = len(self._ssh_credentials) > 0
        return has_creds

    @property
    def identity(self) -> str:
        """
            Returns a string that identifies a device in logs. This property can
            be overridden by custom landscape devices to customize identities in
            logs.
        """
        return self._friendly_id.identity

    @property
    def ipaddr(self):
        """
            The device IP of the device, if available.
        """
        ipaddr = self._resolve_ipaddress()
        return ipaddr

    @property
    def is_watched(self) -> bool:
        """
            A boolean indicating if this device is a watched device.
        """
        return self._is_watched

    @property
    def landscape(self) -> "Landscape":
        """
            Returns a strong reference to the the landscape object
        """
        return self._lscape_ref()

    @property
    def moniker(self) -> str:
        """
            Returns a moniker for this device.
        """
        return self._friendly_id.identity

    @property
    def pivots(self) -> str:
        """
            Returns a set of pivot data points that can be used to
            colate results.
        """
        pivots = (self._friendly_id.identity,)
        return pivots

    @property
    def power(self) -> "LandscapeDeviceExtension":
        """
            The power agent associated with this device.
        """
        return self._power

    @property
    def serial(self) -> "LandscapeDeviceExtension":
        """
            The serial agent associated with this device.
        """
        return self._serial

    @property
    def ssh(self) -> "SshAgent":
        """
            The 'SSH' :class:`SshAgent` attached to this device or None.
        """
        return self._ssh

    @property
    def ssh_credentials(self):
        """
            The list of SSH credentials associated with this device
        """
        return self._ssh_credentials

    @property
    def upnp(self) -> "UpnpRootDevice":
        """
            The 'UPnP' :class:`UpnpRootDevice` attached to this device or None.
        """
        return self._upnp

    def attach_extension(self, ext_type, extension):
        """
            Method called by device coordinators to attach a device extension to a :class:`LandscapeDevice`.
        """
        setattr(self, "_" + ext_type, extension)
        return

    def checkout(self):
        """
            Method that makes it convenient to checkout device.
        """
        self.landscape.checkout_device(self)
        return

    def checkin(self):
        """
            Method that makes it convenient to checkin a device.
        """
        self.landscape.checkin_device(self)
        return

    def get_preferred_command_runner(self) -> ICommandRunner:
        """
            Gets the preferred command runner interface for this device.
        """
        cmd_runner = None

        if self._preferred_command_interface == "ssh":
            if hasattr(self, "ssh") and self.ssh is not None:
                cmd_runner = self.ssh
            elif hasattr(self, "serial") and self.serial is not None:
                cmd_runner = self.serial
        elif self._preferred_command_interface == "serial":
            if hasattr(self, "serial") and self.serial is not None:
                cmd_runner = self.serial
            elif hasattr(self, "ssh") and self.ssh is not None:
                cmd_runner = self.ssh
        else:
            errmsg = "Unknown command runner interface '{}'".format(self._preferred_command_interface)
            raise AKitRuntimeError(errmsg)

        return cmd_runner

    def initialize_features(self):
        """
            Initializes the features of the device based on the feature declarations and the information
            found in the feature config.
        """
        if "features" in self._device_config:
            feature_info = self._device_config["features"]
            for fkey, fval in feature_info.items():
                if fkey == "isolation":
                    self._is_isolated = fval
                elif fkey == "power":
                    self._intitialize_power(fval)
                elif fkey == "serial":
                    self._intitialize_serial(fval)
                elif fkey == "":
                    pass
        return

    def match_using_params(self, match_type, *match_params) -> bool:
        """
            Method that allows you to match :class:`LandscapeDevice` objects by providing a match_type and
            parameters.  The match type is mapped to functions that are registered by device coordinators
            and then the function is called with the match parameters to determine if a device is a match.
        """
        matches = False
        match_func = None
        match_self = None

        self._device_lock.acquire()
        try:
            if match_type in self._match_functions:
                dext_attr, match_func = self._match_functions[match_type]
                match_self = None
                if hasattr(self, dext_attr):
                    match_self = getattr(self, dext_attr)
        finally:
            self._device_lock.release()

        if match_self is not None and match_func is not None:
            matches = match_func(match_self, *match_params)

        return matches

    def set_preferred_command_interface(self, ifname: str):
        """
            Sets the name of the interface that is used as the preferred command runner.

            :param ifname: The name of the preferred commandline interface. (ssh, serial, etc.)
        """
        self._preferred_command_interface = ifname
        return

    def update_full_identifier(self, full_identifier: str):
        """
            Update the full identifier associated with the FriendlyIdentifier for this device.
        """
        self._friendly_id.update_full_identifier(full_identifier)
        return

    def update_match_table(self, match_table: dict):
        """
            Method called  to update the match functions.
        """
        self._device_lock.acquire()
        try:
            self._match_functions.update(match_table)
        finally:
            self._device_lock.release()

        return

    def verify_status(self):
        """
            Verify the status of the specified device.
        """

        status = "Up"
        if self._upnp is not None:
            desc = self._upnp.query_device_description()
            if desc is None:
                status = "Down"

        elif self._ssh:
            cmd_status, _, _ = self._ssh.run_cmd("echo Hello")
            if cmd_status != 0:
                status = "Down"

        return status

    def _intitialize_power(self, power_mapping): # pylint: disable=no-self-use
        """
            Initializes the serial port connectivity for this device.
        """
        self._power = self.landscape.lookup_power_agent(power_mapping)
        return

    def _intitialize_serial(self, serial_mapping): # pylint: disable=no-self-use
        """
            Initializes the serial port connectivity for this device.
        """
        self._serial = self.landscape.lookup_serial_agent(serial_mapping)
        return

    def _resolve_ipaddress(self):
        ipaddr = None
        if self._device_type == "network/upnp" and self.upnp is not None:
            ipaddr = self.upnp.IPAddress
        elif self._device_type == "network/ssh" and self.ssh is not None:
            ipaddr = self.ssh.ipaddr
        return ipaddr

    def _repr_html_(self):
        html_repr_lines = [
            "<h1>LandscapeDevice</h1>",
            "<h2>     type: {}</h2>".format(self._device_type),
            "<h2>    identity: {}</h2>".format(self.identity),
            "<h2>       ip: {}</h2>".format(self.ipaddr)
        ]

        html_repr = os.linesep.join(html_repr_lines)

        return html_repr

    def __repr__(self):

        thisType = type(self)

        ipaddr = self.ipaddr
        devstr = "<{} type={} identity={} ip={} >".format(thisType.__name__, self._device_type, self.identity, ipaddr)

        return devstr

    def __str__(self):
        devstr = repr(self)
        return devstr


"""
.. module:: landscapeintegrationlayer
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the Landscape related classes.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from akit.coupling.basecoupling import BaseCoupling

from akit.exceptions import AKitSemanticError

class LandscapeIntegrationLayer:
    """

    """

    _activated_gate = None

    def __init__(self):
        self._ordered_roles = []

        self._integration_points_registered = {}

        self._integration_point_registration_counter = 0

        super().__init__()

        # We need to wait till we have initialized the landscape configuration
        # layer before we start registering integration points
        lscapeType = type(self)
        lscapeType.landscape_description.register_integration_points(self)

        return

    def register_integration_point(self, role: str, coupling: BaseCoupling):
        """
            This method should be called from the attach_to_environment methods from individual couplings
            in order to register the base level integrations.  Integrations can be hierarchical so it
            is only necessary to register the root level integration couplings, the descendant couplings can
            be called from the root level couplings.

            :param role: The name of a role to assign for a coupling.
            :param coupling: The coupling to register for the associated role.
        """
        lscapeType = type(self)

        lscapeType.landscape_lock.acquire()
        try:
            if role not in self._integration_points_registered:
                self._ordered_roles.append(role)
                self._integration_points_registered[role] = coupling

                self._integration_point_registration_counter += 1
            else:
                raise AKitSemanticError("A coupling with the role %r was already registered." % role) from None
        finally:
            lscapeType.landscape_lock.release()

        return

    def activate_integration(self):
        """
            Called in order to mark the configuration process as complete in order
            for the activation stage to begin and to make the activation level methods
            callable.
        """
        self._landscape_configured = True
        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class ConfigurationManagement1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:ConfigurationManagement:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'ConfigurationManagement1'

    SERVICE_DEFAULT_VARIABLES = {
        "AttributeValuesUpdate": { "data_type": "string", "default": None, "allowed_list": None},
        "ConfigurationUpdate": { "data_type": "string", "default": None, "allowed_list": None},
        "CurrentConfigurationVersion": { "data_type": "ui4", "default": None, "allowed_list": None},
        "InconsistentStatus": { "data_type": "boolean", "default": None, "allowed_list": None},
        "SupportedDataModelsUpdate": { "data_type": "string", "default": None, "allowed_list": None},
        "SupportedParametersUpdate": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_CreateInstance(self, MultiInstanceName, ChildrenInitialization, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CreateInstance action.

            :returns: "InstanceIdentifier", "Status"
        """
        arguments = {
            "MultiInstanceName": MultiInstanceName,
            "ChildrenInitialization": ChildrenInitialization,
        }

        out_params = self.call_action("CreateInstance", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("InstanceIdentifier", "Status",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DeleteInstance(self, InstanceIdentifier, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteInstance action.

            :returns: "Status"
        """
        arguments = {
            "InstanceIdentifier": InstanceIdentifier,
        }

        out_params = self.call_action("DeleteInstance", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Status",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAttributeValuesUpdate(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAttributeValuesUpdate action.

            :returns: "StateVariableValue"
        """
        arguments = { }

        out_params = self.call_action("GetAttributeValuesUpdate", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAttributes(self, Parameters, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAttributes action.

            :returns: "NodeAttributeValueList"
        """
        arguments = {
            "Parameters": Parameters,
        }

        out_params = self.call_action("GetAttributes", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NodeAttributeValueList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetConfigurationUpdate(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetConfigurationUpdate action.

            :returns: "StateVariableValue"
        """
        arguments = { }

        out_params = self.call_action("GetConfigurationUpdate", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCurrentConfigurationVersion(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCurrentConfigurationVersion action.

            :returns: "StateVariableValue"
        """
        arguments = { }

        out_params = self.call_action("GetCurrentConfigurationVersion", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetInconsistentStatus(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetInconsistentStatus action.

            :returns: "StateVariableValue"
        """
        arguments = { }

        out_params = self.call_action("GetInconsistentStatus", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetInstances(self, StartingNode, SearchDepth, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetInstances action.

            :returns: "Result"
        """
        arguments = {
            "StartingNode": StartingNode,
            "SearchDepth": SearchDepth,
        }

        out_params = self.call_action("GetInstances", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSelectedValues(self, StartingNode, Filter, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSelectedValues action.

            :returns: "ParameterValueList"
        """
        arguments = {
            "StartingNode": StartingNode,
            "Filter": Filter,
        }

        out_params = self.call_action("GetSelectedValues", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ParameterValueList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSupportedDataModels(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSupportedDataModels action.

            :returns: "SupportedDataModels"
        """
        arguments = { }

        out_params = self.call_action("GetSupportedDataModels", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SupportedDataModels",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSupportedDataModelsUpdate(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSupportedDataModelsUpdate action.

            :returns: "StateVariableValue"
        """
        arguments = { }

        out_params = self.call_action("GetSupportedDataModelsUpdate", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSupportedParameters(self, StartingNode, SearchDepth, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSupportedParameters action.

            :returns: "Result"
        """
        arguments = {
            "StartingNode": StartingNode,
            "SearchDepth": SearchDepth,
        }

        out_params = self.call_action("GetSupportedParameters", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSupportedParametersUpdate(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSupportedParametersUpdate action.

            :returns: "StateVariableValue"
        """
        arguments = { }

        out_params = self.call_action("GetSupportedParametersUpdate", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StateVariableValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetValues(self, Parameters, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetValues action.

            :returns: "ParameterValueList"
        """
        arguments = {
            "Parameters": Parameters,
        }

        out_params = self.call_action("GetValues", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ParameterValueList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetAttributes(self, NodeAttributeValueList, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAttributes action.

            :returns: "Status"
        """
        arguments = {
            "NodeAttributeValueList": NodeAttributeValueList,
        }

        out_params = self.call_action("SetAttributes", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Status",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetValues(self, ParameterValueList, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetValues action.

            :returns: "Status"
        """
        arguments = {
            "ParameterValueList": ParameterValueList,
        }

        out_params = self.call_action("SetValues", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Status",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""
.. module:: orm
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the ORM associated with the akit database storage

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import enum
import json

from sqlalchemy import BigInteger, Column, DateTime, Enum, Float, String, Text, VARCHAR, ForeignKey, TEXT
from sqlalchemy import inspect
from sqlalchemy import MetaData
from sqlalchemy.types import JSON

from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy_utils.types.uuid import UUIDType


class SerializableModel:
    """
        Mixin style class that adds serialization to data model objects.
    """

    def to_dict(self):
        """
            Iterates the formal data attributes of a model and outputs a dictionary
            with the data based on the model.
        """
        dval = {}

        model = type(self)
        mapper = inspect(model)
        for col in mapper.attrs:
            col_key = col.key
            dval[col_key] = str(getattr(self, col_key))

        return dval


    def to_json(self, indent=4):
        """
            Iterates the formal data attributes of a model and creates a dictionary
            with the data based on the model, then converts the dictionary to a
            JSON string.
        """
        model_dict = self.to_dict()
        json_str = json.dumps(model_dict, indent=indent)
        return json_str

AutomationBase = declarative_base()


class RepositoryType(enum.Enum):
    """
        An enumeration that indicates the JobType of a WorkQueue item.  This indicates if a
        work item is a global work item available for execution on any qualified work site or
        if it is a local work item which is meant to execute locally.
    """
    GitHub = 1

class WorkQueueJobType(enum.Enum):
    """
        An enumeration that indicates the JobType of a WorkQueue item.  This indicates if a
        work item is a global work item available for execution on any qualified work site or
        if it is a local work item which is meant to execute locally.
    """
    Local = 1
    Global = 2

class WorkQueueState(enum.Enum):
    """
        An enumeration that indicates the JobType of a WorkQueue item.  This indicates if a
        work item is a global work item available for execution on any qualified work site or
        if it is a local work item which is meant to execute locally.
    """
    Enabled = 1
    Disabled = 2

class Repository(AutomationBase, SerializableModel):
    """
        A data modelf for storing repository information.
    """

    __tablename__ = "repository"

    id = Column('repo_id', BigInteger, primary_key=True)
    name = Column('repo_name', VARCHAR(128), nullable=False)
    url = Column('repo_repository', VARCHAR(2048), nullable=False)
    repotype = Column('repo_type', Enum(RepositoryType), nullable=False)
    repoconf = Column('repo_conf', TEXT, nullable=False)

class Branch(AutomationBase, SerializableModel):
    """
        A data model for storing source branch information.
    """

    __tablename__ = "branch"

    id = Column('branch_id', BigInteger, primary_key=True)
    name = Column('branch_name', VARCHAR(128), nullable=False)
    owner = Column('branch_owner', VARCHAR(128), nullable=False)

    repo_id = Column('repo_id', BigInteger, ForeignKey(Repository.id))

class BuildResult(AutomationBase, SerializableModel):
    """
        A data model for storing build information.
    """

    __tablename__ = "build_result"

    id = Column('build_id', BigInteger, primary_key=True)
    name = Column('build_name', VARCHAR(128), nullable=False)
    start = Column('build_start', DateTime, nullable=False)
    stop = Column('build_stop', DateTime, nullable=False)
    change = Column('build_change', VARCHAR(128), nullable=False)
    
    branch_id = Column('branch_id', BigInteger, ForeignKey(Branch.id))

class BuildByProduct(AutomationBase, SerializableModel):
    """
        A data mode for linking build byproducts to build
    """
    __tablename__ = "build_byproduct"

    id = Column('byprod_id', BigInteger, primary_key=True)
    product = Column('byprod_product', VARCHAR(128), nullable=False)
    arch = Column('byprod_arch', VARCHAR(128), nullable=False)
    buildno = Column('byprod_buildno', VARCHAR(128), nullable=False)
    flavor = Column('byprod_flavor', VARCHAR(128), nullable=False)
    result_url = Column('byprod_result_url', VARCHAR(2048), nullable=False) 
    
    bldjob_id = Column('bldjob_id', BigInteger, ForeignKey(BuildResult.id))

class User(AutomationBase, SerializableModel):
    """
        A data model for the minimal amount of User information to store
        for relating automation data to an enterprise user.
    """

    __tablename__ = 'user'

    id = Column('user_id', VARCHAR(40), primary_key=True)
    firstName = Column('user_firstName', VARCHAR(40), nullable=False)
    lastName = Column('user_lastName', VARCHAR(40), nullable=False)
    email = Column('user_email', VARCHAR(128), nullable=False)
    login = Column('user_login', VARCHAR(128), nullable=False)
    ownertag = Column('user_ownertag', VARCHAR(40), nullable=True)

class WorkQueue(AutomationBase, SerializableModel):
    """
        A data model for a
    """
    __tablename__ = 'work_queue'
    
    id = Column('queue_id', BigInteger, primary_key=True, autoincrement=True)
    name = Column('queue_name', String(128), nullable=False)
    description = Column('queue_desc', TEXT, nullable=False)
    state = Column('queue_state', Enum(WorkQueueState), nullable=False)

class WorkPacket(AutomationBase, SerializableModel):
    """
        A data model for a WorkPacket(s) and the work items that are part of a work queue.
    """
    __tablename__ = 'work_packet'
    
    id = Column('wkpack_id', BigInteger, primary_key=True, autoincrement=True)

    jtype = Column('wkpk_jtype', Enum(WorkQueueJobType), nullable=False)
    title =  Column('wkpk_title', String(1024), nullable=False)
    description = Column('wkpk_description', Text, nullable=False)
    branch =  Column('wkpk_branch', String(1024), nullable=True)
    build =  Column('wkpk_build', String(1024), nullable=True)
    flavor =  Column('wkpk_flavor', String(1024), nullable=True)
    ownertag = Column('wkpk_ownertag', VARCHAR(40), nullable=True)
    added = Column('wkpk_added', DateTime, nullable=False)
    start = Column('wkpk_start', DateTime, nullable=True)
    stop = Column('wkpk_stop', DateTime, nullable=True)
    progress = Column('wkpk_progress', Float, default=0.0)
    status = Column('wkpk_status', String(50), nullable=False)
    packet = Column('wkpk_packet', TEXT, nullable=False)

    queue_id = Column('queue_id', BigInteger, ForeignKey(WorkQueue.id), nullable=False)
    user_id = Column('user_id', VARCHAR(40), ForeignKey(User.id), nullable=False)


class FeatureGroup(AutomationBase, SerializableModel):
    """
        A data model for storing information about feature groups.
    """
    __tablename__ = 'feature_group'
    
    id = Column('lsscan_id', BigInteger, primary_key=True)
    name = Column('fg_name', VARCHAR(256), nullable=False)
    description = Column('fg_description', TEXT, nullable=False)

class Landscape(AutomationBase, SerializableModel):
    """
        A data model that describes a test landscape.
    """
    __tablename__ = 'landscape'
    
    id = Column('lsdesc_id', BigInteger, primary_key=True)
    name =  Column('lsdesc_name', VARCHAR(1024), nullable=False)
    detail = Column('lsdesc_detail', JSON, nullable=False)

class LandscapeScan(AutomationBase, SerializableModel):
    """
        A data model that describes the results of a test landscape scan.
    """
    __tablename__ = 'landscape_scan'
    
    id = Column('lsscan_id', BigInteger, primary_key=True)
    name =  Column('lsscan_name', VARCHAR(1024), nullable=False)
    detail = Column('lsscan_detail', JSON, nullable=False)

    lscape_id = Column('lsdesc_id', BigInteger, ForeignKey(Landscape.id))

class TestJob(AutomationBase, SerializableModel):
    """
        A data model for storing information about test jobs and
        user and feature group associations with a given test job.
    """
    __tablename__ = 'test_job'
    
    id = Column('job_id', VARCHAR(40), primary_key=True)
    title =  Column('job_title', VARCHAR(1024), nullable=False)
    description = Column('job_description', Text, nullable=False)

    owner_fg_id = Column('owner_fg_id', BigInteger, ForeignKey(FeatureGroup.id), nullable=True)

class TestJobRun(AutomationBase, SerializableModel):
    """
        A data model for a TestJob run.
    """

    __tablename__ = 'test_job_run'
    
    id = Column('tjr_id', BigInteger, primary_key=True)
    instance = Column('tjr_instance', UUIDType, nullable=False)
    branch =  Column('tjr_branch', VARCHAR(1024), nullable=True)
    build =  Column('tjr_build', VARCHAR(1024), nullable=True)
    flavor =  Column('tjr_flavor', VARCHAR(1024), nullable=True)
    ownertag = Column('wkpk_ownertag', VARCHAR(40), nullable=True)
    start = Column('tjr_start', DateTime, nullable=False)
    stop = Column('tjr_stop', DateTime, nullable=True)
    detail = Column('tjr_detail', JSON, nullable=True)

    job_id = Column('job_id', VARCHAR(40), ForeignKey(TestJob.id), nullable=False)
    user_id = Column('user_id', VARCHAR(40), ForeignKey(User.id), nullable=False)

    lscape_id = Column('lsdesc_id', BigInteger, ForeignKey(Landscape.id), nullable=True)
    lsscan_id = Column('lsscan_id', BigInteger, ForeignKey(LandscapeScan.id), nullable=True)
    fg_id = Column('fg_id', BigInteger, ForeignKey(FeatureGroup.id), nullable=True)

class TestResult(AutomationBase, SerializableModel):
    """
        A data model for a TestResult node that is part of a test result tree.
    """
    __tablename__ = 'test_result'
    
    id = Column('tstr_id', BigInteger, primary_key=True)
    name =  Column('tstr_name', VARCHAR(1024), nullable=False)
    extname = Column('tstr_extname', VARCHAR(1024), nullable=True)
    parameters = Column('tstr_parameters', Text, nullable=True)
    instance = Column('tstr_instance', UUIDType, nullable=False)
    parent = Column('tstr_parent', UUIDType, nullable=True)
    rtype = Column('tstr_rtype', String(50), nullable=False)
    result = Column('tstr_result', String(50), nullable=False)
    start = Column('tstr_start', DateTime, nullable=False)
    stop = Column('tstr_stop', DateTime, nullable=True)
    detail = Column('tstr_detail', JSON, nullable=True)

    testjob_id = Column('tj_id', VARCHAR(40), ForeignKey(TestJob.id))

class TestResultContainer(AutomationBase, SerializableModel):
    """
        A data model for a TestResultContainer node that is part of a test result tree.  The
        TestResultContainer node serves as a parent and container for individual result based
        nodes.
    """
    __tablename__ = 'test_result_container'
    
    id = Column('tstrcont_id', BigInteger, primary_key=True)
    name =  Column('tstrcont_name', VARCHAR(1024), nullable=False)
    instance = Column('tstrcont_instance', UUIDType, nullable=False)
    parent = Column('tstrcont_parent', UUIDType, nullable=True)
    rtype = Column('tstrcont_rtype', String(50), nullable=False)

    testjob_id = Column('tj_id', VARCHAR(40), ForeignKey(TestJob.id))

class FeatureGroupMembership(AutomationBase, SerializableModel):
    """
        A data model for associating FeatureGroup(s) to Users(s)
    """
    __tablename__ = "feature_group_membership"
    
    fg_id = Column('fg_id', BigInteger, ForeignKey(FeatureGroup.id), primary_key=True)
    user_id = Column('user_id', VARCHAR(40), ForeignKey(User.id), primary_key=True)

    join_date = Column('fgm_join_date', DateTime, nullable=False)
    departure_date = Column('fgm_departure_date', DateTime, nullable=True)




"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class TemperatureSensor1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:TemperatureSensor:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'TemperatureSensor1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "Application": { "data_type": "string", "default": "Room", "allowed_list": "['Room', 'Outdoor', 'Pipe', 'AirDuct']"},
        "CurrentTemperature": { "data_type": "i4", "default": "2000", "allowed_list": None},
        "Name": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetApplication(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetApplication action.

            :returns: "CurrentApplication"
        """
        arguments = { }

        out_params = self.call_action("GetApplication", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentApplication",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCurrentTemperature(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCurrentTemperature action.

            :returns: "CurrentTemp"
        """
        arguments = { }

        out_params = self.call_action("GetCurrentTemperature", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTemp",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetName(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetName action.

            :returns: "CurrentName"
        """
        arguments = { }

        out_params = self.call_action("GetName", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentName",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetApplication(self, NewApplication, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetApplication action.
        """
        arguments = {
            "NewApplication": NewApplication,
        }

        self.call_action("SetApplication", arguments=arguments, aspects=aspects)

        return

    def action_SetName(self, NewName, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetName action.
        """
        arguments = {
            "NewName": NewName,
        }

        self.call_action("SetName", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class ExternalActivity1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:ExternalActivity:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'ExternalActivity1'

    SERVICE_DEFAULT_VARIABLES = {
        "ButtonName": { "data_type": "string", "default": "All", "allowed_list": "['All', 'Scan']"},
        "DisplayString": { "data_type": "string", "default": None, "allowed_list": None},
        "DisplayStringSize": { "data_type": "ui4", "default": None, "allowed_list": None},
        "Duration": { "data_type": "i4", "default": None, "allowed_list": None},
        "RegistrationID": { "data_type": "ui4", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "Activity": { "data_type": "string", "default": None, "allowed_list": None},
        "AvailableRegistrations": { "data_type": "boolean", "default": "1", "allowed_list": None},
    }

    def action_Register(self, ButtonNameIn, DisplayStringIn, DurationIn, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Register action.

            :returns: "ActualDurationOut", "RegistrationIDOut"
        """
        arguments = {
            "ButtonNameIn": ButtonNameIn,
            "DisplayStringIn": DisplayStringIn,
            "DurationIn": DurationIn,
        }

        out_params = self.call_action("Register", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ActualDurationOut", "RegistrationIDOut",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""
.. module:: aspects
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains the :class:`Aspects` class and the constants used to provide aspect behaviors.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional, Union

from enum import IntEnum

from logging import Logger

from akit.timeouts import TimeoutContext
from akit.xlogging.foundations import getAutomatonKitLogger

class ActionPattern(IntEnum):
    """
        A action pattern to use when using apspects to change the behavior of an API.
    """
    SINGLE_CALL = 0
    SINGLE_CONNECTED_CALL = 1
    DO_UNTIL_SUCCESS = 2
    DO_WHILE_SUCCESS = 3
    DO_UNTIL_CONNECTION_FAILURE = 4

class LoggingPattern:
    """
        A logging pattern to use when logging commands.
    """
    ALL_RESULTS = 0
    FAILURE_ONLY = 1
    SUCCESS_ONLY = 2


DEFAULT_COMPLETION_TIMEOUT = 600
DEFAULT_COMPLETION_INTERVAL = 10

DEFAULT_INACTIVITY_TIMEOUT = 600
DEFAULT_INACTIVITY_INTERVAL = .5

DEFAULT_MONITOR_DELAY = 30

DEFAULT_RETRY_LOGGING_INTERVAL = 5

DEFAULT_LOGGING_PATTERN = LoggingPattern.ALL_RESULTS

DEFAULT_MUST_CONNECT = False

DEFAULT_ALLOWED_ERROR_CODES: List[int] = []

class Aspects:
    """
        Aspects are utilized with the interop APIs and agents such as the :class:`SSHAgent` class in order
        to modify the behavior of APIs with respect to retry parameter such as timeout, interval, looping patterns
        logging, etc.  The aspects object provides a way to package this common criteria into a single parameter
        or constant you can  pass to multiple APIs
    """

    def __init__(self, action_pattern: ActionPattern = ActionPattern.SINGLE_CALL,
                       completion_timeout: float = DEFAULT_COMPLETION_TIMEOUT,
                       completion_interval: float = DEFAULT_COMPLETION_INTERVAL,
                       inactivity_timeout: float = DEFAULT_INACTIVITY_TIMEOUT,
                       inactivity_interval: float = DEFAULT_INACTIVITY_INTERVAL,
                       monitor_delay: float = DEFAULT_MONITOR_DELAY,
                       logging_pattern: LoggingPattern = DEFAULT_LOGGING_PATTERN,
                       retry_logging_interval: int = DEFAULT_RETRY_LOGGING_INTERVAL,
                       allowed_error_codes: List[int] = DEFAULT_ALLOWED_ERROR_CODES,
                       must_connect: bool = DEFAULT_MUST_CONNECT,
                       logger: Optional[Logger]=None):
        """
            Creates an :class:`Aspects` instance package.

            :param action_pattern: The :class:`ActionPattern` that the API should exhibit such as
                                   SINGULAR, DO_UNTIL_SUCCESS, DO_WHILE_SUCCESS
            :param completion_timeout: The time in seconds as a float that is the max time before
                                       timeout for the activity to complete.
            :param completion_interval: The time in seconds as a float that is waited before reattempting
                                        an activity.
            :param inactivity_timeout: The time in seconds as a float that is the max time before timeout
                                       that is waited before a :class:`TimeoutError` is raised due to inactivity.
            :param inactivity_interval: The time in seconds as a float that is waited before reattempting an activity.
            :param retry_logging_interval: The logging interval for retry loops.
            :param allowed_error_codes: A list of error codes that allow retry
            :param must_connect: Used to indicate that connection failures should be raised.
        """
        self.action_pattern = action_pattern
        self.completion_timeout = completion_timeout
        self.completion_interval = completion_interval
        self.inactivity_timeout = inactivity_timeout
        self.inactivity_interval = inactivity_interval
        self.monitor_delay = monitor_delay
        self.logging_pattern = logging_pattern
        self.retry_logging_interval = retry_logging_interval
        self.allowed_error_codes = allowed_error_codes
        self.must_connect=must_connect

        if logger is None:
            self.logger = getAutomatonKitLogger()
        else:
            self.logger = logger

        return

class AspectsCmd(Aspects):
    """
    """
    def __init__(self, expected_status: Union[int, List[int]]=0,
                       user: Optional[str]=None,
                       pty_params: Optional[dict]=None,
                       action_pattern: ActionPattern = ActionPattern.SINGLE_CALL,
                       completion_timeout: float = DEFAULT_COMPLETION_TIMEOUT,
                       completion_interval: float = DEFAULT_COMPLETION_INTERVAL,
                       inactivity_timeout: float = DEFAULT_INACTIVITY_TIMEOUT,
                       inactivity_interval: float = DEFAULT_INACTIVITY_INTERVAL,
                       monitor_delay: float = DEFAULT_MONITOR_DELAY,
                       logging_pattern: LoggingPattern = DEFAULT_LOGGING_PATTERN,
                       retry_logging_interval: int = DEFAULT_RETRY_LOGGING_INTERVAL,
                       allowed_error_codes: List[int] = DEFAULT_ALLOWED_ERROR_CODES,
                       logger: Optional[Logger]=None):
        
        Aspects.__init__(self, action_pattern=action_pattern, completion_timeout=completion_timeout,
                            completion_interval=completion_interval, inactivity_timeout=inactivity_timeout,
                            inactivity_interval=inactivity_interval, monitor_delay=monitor_delay,
                            logging_pattern=logging_pattern, retry_logging_interval=retry_logging_interval,
                            allowed_error_codes=allowed_error_codes, logger=logger)
        
        self.expected_status = expected_status
        self.user = user
        self.pty_params = pty_params
        return

class AspectsUPnP(Aspects):
    """
    """
    def __init__(self, action_pattern: ActionPattern = ActionPattern.SINGLE_CONNECTED_CALL,
                       completion_timeout: float = DEFAULT_COMPLETION_TIMEOUT,
                       completion_interval: float = DEFAULT_COMPLETION_INTERVAL,
                       inactivity_timeout: float = DEFAULT_INACTIVITY_TIMEOUT,
                       inactivity_interval: float = DEFAULT_INACTIVITY_INTERVAL,
                       monitor_delay: float = DEFAULT_MONITOR_DELAY,
                       logging_pattern: LoggingPattern = DEFAULT_LOGGING_PATTERN,
                       retry_logging_interval: int = DEFAULT_RETRY_LOGGING_INTERVAL,
                       allowed_error_codes: List[int] = DEFAULT_ALLOWED_ERROR_CODES,
                       must_connect: bool = DEFAULT_MUST_CONNECT,
                       logger: Optional[Logger]=None):
        
        Aspects.__init__(self, action_pattern=action_pattern, completion_timeout=completion_timeout,
                            completion_interval=completion_interval, inactivity_timeout=inactivity_timeout,
                            inactivity_interval=inactivity_interval, monitor_delay=monitor_delay,
                            logging_pattern=logging_pattern, retry_logging_interval=retry_logging_interval,
                            allowed_error_codes=allowed_error_codes, must_connect=must_connect, logger=logger)
        return


DEFAULT_ASPECTS = Aspects()
DEFAULT_CMD_ASPECTS = AspectsCmd()
DEFAULT_UPNP_ASPECTS = AspectsUPnP()



"""
.. module:: serialcoordinator
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the SerialCoordinator which is used for managing serial activity services.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import TYPE_CHECKING

from akit.exceptions import AKitConfigurationError
from akit.interop.coordinators.coordinatorbase import CoordinatorBase

from akit.interop.agents.serialagents import TcpSerialAgent

if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class SerialCoordinator(CoordinatorBase):

    def __init__(self, lscape: "Landscape", *args, **kwargs):
        super(SerialCoordinator, self).__init__(lscape, *args, **kwargs)
        return

    def _initialize(self, *_args, **_kwargs):
        """
            Called by the CoordinatorBase constructor to perform the one time initialization
            of the coordinator Singleton of a given type.
        """
        self._serial_config = {}
        for scfg in self._coord_config:
            cfgname = scfg["name"]
            self._serial_config[cfgname] = scfg

        self._serial_agent = {}
        return

    def lookup_agent(self, serial_mapping: dict) -> TcpSerialAgent:
        """
            Looks up a serial agent by serial mapping.
        """
        serial_agent = None

        interface_name = serial_mapping["name"]
        attachment_point = serial_mapping["port"]

        lscape = self.landscape

        if interface_name in self._serial_config:
            serial_config = self._serial_config[interface_name]
            serialType = serial_config["serialType"]
            if serialType == "network/tcp":
                host = serial_config["host"]
                ports_table = serial_config["ports"]
                port = ports_table[attachment_point]

                serial_agent = TcpSerialAgent(host, port)

                self._serial_agent[serial_mapping] = serial_agent
            else:
                errmsg = "Invalid serialType=%s for serial interface %r." % (serialType, interface_name)
                raise AKitConfigurationError(errmsg) from None
        else:
            errmsg = "Failure to lookup serial interface %r." % interface_name
            raise AKitConfigurationError(errmsg) from None

        return serial_agent



"""
.. module:: akit.interop.upnp.extensions.sonos.zps23
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the a Upnp device for a Sonos Zps23.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from akit.interop.upnp.devices.upnprootdevice import UpnpRootDevice

from akit.interop.upnp.xml.upnpdevice1 import UpnpDevice1Device

class SonosUpnpDevice1Device(UpnpDevice1Device):
    """
    """

    @property
    def ampOnTime(self):
        rtnval = self._find_value("ampOnTime", namespaces=self._namespaces)
        return rtnval

    @property
    def apiVersion(self):
        rtnval = self._find_value("apiVersion", namespaces=self._namespaces)
        return rtnval

    @property
    def bassExtension(self):
        rtnval = self._find_value("bassExtension", namespaces=self._namespaces)
        return rtnval

    @property
    def displayName(self):
        rtnval = self._find_value("displayName", namespaces=self._namespaces)
        return rtnval

    @property
    def displayVersion(self):
        rtnval = self._find_value("displayVersion", namespaces=self._namespaces)
        return rtnval

    @property
    def extraVersion(self):
        rtnval = self._find_value("extraVersion", namespaces=self._namespaces)
        return rtnval

    @property
    def feature1(self):
        rtnval = self._find_value("feature1", namespaces=self._namespaces)
        return rtnval

    @property
    def feature2(self):
        rtnval = self._find_value("feature2", namespaces=self._namespaces)
        return rtnval

    @property
    def feature3(self):
        rtnval = self._find_value("feature3", namespaces=self._namespaces)
        return rtnval

    @property
    def flash(self):
        rtnval = self._find_value("flash", namespaces=self._namespaces)
        return rtnval

    @property
    def hardwareVersion(self):
        rtnval = self._find_value("hardwareVersion", namespaces=self._namespaces)
        return rtnval

    @property
    def internalSpeakerSize(self):
        rtnval = self._find_value("internalSpeakerSize", namespaces=self._namespaces)
        return rtnval

    @property
    def legacyCompatibleVersion(self):
        rtnval = self._find_value("legacyCompatibleVersion", namespaces=self._namespaces)
        return rtnval

    @property
    def memory(self):
        rtnval = self._find_value("memory", namespaces=self._namespaces)
        return rtnval

    @property
    def minApiVersion(self):
        rtnval = self._find_value("minApiVersion", namespaces=self._namespaces)
        return rtnval

    @property
    def minCompatibleVersion(self):
        rtnval = self._find_value("minCompatibleVersion", namespaces=self._namespaces)
        return rtnval

    @property
    def retailMode(self):
        rtnval = self._find_value("retailMode", namespaces=self._namespaces)
        return rtnval

    @property
    def roomName(self):
        rtnval = self._find_value("roomName", namespaces=self._namespaces)
        return rtnval

    @property
    def satGainOffset(self):
        rtnval = self._find_value("satGainOffset", namespaces=self._namespaces)
        return rtnval

    @property
    def seriesid(self):
        rtnval = self._find_value("seriesid", namespaces=self._namespaces)
        return rtnval

    @property
    def softwareVersion(self):
        rtnval = self._find_value("softwareVersion", namespaces=self._namespaces)
        return rtnval

    @property
    def swGen(self):
        rtnval = self._find_value("swGen", namespaces=self._namespaces)
        return rtnval

    @property
    def variant(self):
        rtnval = self._find_value("variant", namespaces=self._namespaces)
        return rtnval

    @property
    def zoneType(self):
        rtnval = self._find_value("zoneType", namespaces=self._namespaces)
        return rtnval


class SonosDevice(UpnpRootDevice):
    """
    """

    MANUFACTURER = "SonosInc"

    def __init__(self, manufacturer: str, modelNumber: str, modelDescription: str):
        super(SonosDevice, self).__init__(manufacturer, modelNumber, modelDescription)
        return


